from json import JSONDecodeError
from typing import Any
from jwt import decode
import logging

from pydantic import ConfigDict
from authlib.integrations.starlette_client.apps import StarletteOAuth2App
from httpx import AsyncClient

from oidc_test.settings import AuthProviderSettings, ResourceProvider, Resource, settings
from oidc_test.models import User

logger = logging.getLogger("oidc-test")


class NoPublicKey(Exception):
    pass


class Provider(AuthProviderSettings):
    # To allow authlib_client as StarletteOAuth2App
    model_config = ConfigDict(arbitrary_types_allowed=True)  # type:ignore

    authlib_client: StarletteOAuth2App = StarletteOAuth2App(None)
    info: dict[str, Any] = {}
    unknown_auth_user: User
    logout_with_id_token_hint: bool = True

    def decode(self, token: str, verify_signature: bool | None = None) -> dict[str, Any]:
        """Decode the token with signature check"""
        if self.public_key is None:
            raise NoPublicKey
        if verify_signature is None:
            verify_signature = self.skip_verify_signature
        if settings.debug_token:
            decoded = decode(
                token,
                self.public_key,
                algorithms=[self.signature_alg],
                audience=["account", "oidc-test", "oidc-test-web"],
                options={
                    "verify_signature": False,
                    "verify_aud": False,
                },  # not settings.insecure.skip_verify_signature},
            )
            logger.debug(str(decoded))
        return decode(
            token,
            self.public_key,
            algorithms=[self.signature_alg],
            audience=["account", "oidc-test", "oidc-test-web"],
            options={
                "verify_signature": verify_signature,
            },  # not settings.insecure.skip_verify_signature},
        )

    async def get_info(self):
        # Get the public key:
        async with AsyncClient() as client:
            public_key: str | None = None
            if self.info_url is not None:
                try:
                    provider_info = await client.get(self.info_url)
                except Exception:
                    raise NoPublicKey
                try:
                    self.info = provider_info.json()
                except JSONDecodeError:
                    raise NoPublicKey
                if "public_key" in self.info:
                    # For Keycloak
                    try:
                        public_key = str(self.info["public_key"])
                    except KeyError:
                        raise NoPublicKey
                elif "keys" in self.info:
                    # For Forgejo/Gitea
                    try:
                        public_key = str(self.info["keys"][0]["n"])
                    except KeyError:
                        raise NoPublicKey
            if self.public_key_url is not None:
                resp = await client.get(self.public_key_url)
                public_key = resp.text
            if public_key is None:
                raise NoPublicKey
            self.public_key = "\n".join(
                ["-----BEGIN PUBLIC KEY-----", public_key, "-----END PUBLIC KEY-----"]
            )

    def get_session_key(self, userinfo):
        return userinfo[self.session_key]

    def get_resource(self, resource_name: str) -> Resource:
        return [
            resource for resource in self.resources if resource.resource_name == resource_name
        ][0]

    def get_resource_url(self, resource_name: str) -> str:
        return self.url + self.get_resource(resource_name).url

    def get_resource_provider(self, resource_provider_id: str) -> ResourceProvider:
        return [
            provider
            for provider in self.resource_providers
            if provider.id == resource_provider_id
        ][0]
