from typing import Union, Annotated
from functools import wraps
import logging

from fastapi import HTTPException, Request, Depends, status
from fastapi.security import OAuth2PasswordBearer
from authlib.integrations.starlette_client import OAuth, OAuthError, StarletteOAuth2App
from authlib.oauth2.rfc6749 import OAuth2Token
from jwt import ExpiredSignatureError, InvalidKeyError, DecodeError, PyJWTError

from oidc_test.auth.provider import Provider
from oidc_test.models import User
from oidc_test.database import db, TokenNotInDb, UserNotInDB
from oidc_test.settings import settings
from oidc_test.auth_providers import providers

logger = logging.getLogger("oidc-test")


async def fetch_token(name, request):
    assert name is not None
    assert request is not None
    logger.warning("TODO: fetch_token")
    ...
    # if name in oidc_providers:
    #    model = OAuth2Token
    # else:
    #    model = OAuthToken

    # token = model.find(name=name, user=request.user)
    # return token.to_token()


async def update_token(
    provider_id,
    token,
    refresh_token: str | None = None,
    access_token: str | None = None,
):
    """Update the token in the database"""
    provider = providers[provider_id]
    sid: str = provider.get_session_key(provider.decode(token["id_token"]))
    item = await db.get_token(provider, sid)
    # update old token
    item["access_token"] = token["access_token"]
    item["refresh_token"] = token["refresh_token"]
    item["id_token"] = token["id_token"]
    item["expires_at"] = token["expires_at"]
    logger.info(f"Token {sid} refreshed")
    # It's a fake db and only in memory, so there's nothing to save
    # await item.save()


def init_providers():
    """Add oidc providers to authlib from the settings
    and build the providers dict"""
    for provider_settings in settings.auth.providers:
        provider_settings_dict = provider_settings.model_dump()
        # Add an anonymous user, that cannot be identified but has provided a valid access token
        provider_settings_dict["unknown_auth_user"] = User(
            sub="", auth_provider_id=provider_settings.id
        )
        provider = Provider(**provider_settings_dict)
        if provider.disabled:
            logger.info(f"{provider_settings.name} is disabled, skipping")
        else:
            authlib_oauth.register(
                name=provider.id,
                server_metadata_url=provider.openid_configuration,
                client_kwargs={
                    "scope": " ".join(
                        ["openid", "email", "offline_access", "profile"]
                        + provider.resource_provider_scopes
                    ),
                },
                client_id=provider.client_id,
                client_secret=provider.client_secret,
                api_base_url=provider.url,
                # For PKCE (not implemented yet):
                # code_challenge_method="S256",
                fetch_token=fetch_token,
                update_token=update_token,
                # client_id="some-client-id", # if enabled, authlib will also check that the access token belongs to this client id (audience)
            )
            provider.authlib_client = getattr(authlib_oauth, provider.id)
        providers[provider.id] = provider


authlib_oauth = OAuth(cache=None, fetch_token=fetch_token, update_token=update_token)
oauth2_scheme = OAuth2PasswordBearer(tokenUrl="token")
oauth2_scheme_optional = OAuth2PasswordBearer(tokenUrl="token", auto_error=False)


def get_auth_provider_client_or_none(request: Request) -> StarletteOAuth2App | None:
    """Return the oidc_provider from a request object, from the session.
    It can be used in Depends()"""
    if (auth_provider_id := request.session.get("auth_provider_id")) is None:
        return
    return getattr(authlib_oauth, str(auth_provider_id), None)


def get_auth_provider_client(request: Request) -> StarletteOAuth2App:
    if (oidc_provider := get_auth_provider_client_or_none(request)) is None:
        raise HTTPException(status.HTTP_401_UNAUTHORIZED, "No such provider")
    else:
        return oidc_provider


def get_auth_provider_or_none(request: Request) -> Provider | None:
    """Return the oidc_provider settings from a request object, from the session.
    It can be used in Depends()"""
    if (auth_provider_id := request.session.get("auth_provider_id")) is None:
        return
    return providers.get(auth_provider_id)


def get_auth_provider(request: Request) -> Provider:
    if (provider := get_auth_provider_or_none(request)) is None:
        raise HTTPException(status.HTTP_401_UNAUTHORIZED, "No such provider")
    return provider


async def get_current_user(request: Request) -> User:
    """Get the current user from a request object.
    Also validates the token expiration time.
    ... TODO: complete about refresh token
    """
    if (user_sub := request.session.get("user_sub")) is None:
        raise HTTPException(status.HTTP_401_UNAUTHORIZED)
    token = await get_token_from_session(request)
    user = await db.get_user(user_sub)
    ## Check if the token is expired
    if token.is_expired():
        provider = get_auth_provider(request=request)
        ## Ask a new refresh token from the provider
        logger.info(f"Token expired for user {user.name}")
        try:
            userinfo = await provider.authlib_client.fetch_access_token(
                refresh_token=token.get("refresh_token")
            )
            assert userinfo is not None
        except OAuthError as err:
            logger.exception(err)
            # raise HTTPException(
            #    status.HTTP_401_UNAUTHORIZED, "Token expired, cannot refresh"
            # )

    return user


async def get_token_from_session_or_none(request: Request) -> OAuth2Token | None:
    """Return the auth token from the session or None.
    Can be used in Depends()"""
    try:
        return await get_token_from_session(request)
    except HTTPException:
        return None


async def get_token_from_session(request: Request) -> OAuth2Token:
    """Return the token from the session.
    Can be used in Depends()"""
    try:
        provider = providers[request.session.get("auth_provider_id", "")]
    except KeyError:
        request.session.pop("auth_provider_id", None)
        request.session.pop("user_sub", None)
        raise HTTPException(status.HTTP_401_UNAUTHORIZED, "Invalid provider")
    try:
        return await db.get_token(
            provider,
            request.session.get("sid"),
        )
    except (TokenNotInDb, InvalidKeyError, DecodeError) as err:
        raise HTTPException(status.HTTP_401_UNAUTHORIZED, err.__class__.__name__)


async def get_current_user_or_none(request: Request) -> User | None:
    """Return the user from a request object, from the session.
    It can be used in Depends()"""
    try:
        return await get_current_user(request)
    except HTTPException:
        return None


def hasrole(required_roles: Union[str, list[str]] = []):
    """Decorator for RBAC permissions"""
    required_roles_set: set[str]
    if isinstance(required_roles, str):
        required_roles_set = set([required_roles])
    else:
        required_roles_set = set(required_roles)

    def decorator(func):
        @wraps(func)
        async def wrapper(request=None, *args, **kwargs):
            if request is None:
                raise HTTPException(
                    500,
                    "Functions decorated with hasrole must have a request:Request argument",
                )
            user: User = await get_current_user(request)
            if not any(required_roles_set.intersection(user.roles_as_set)):
                raise HTTPException(status.HTTP_401_UNAUTHORIZED)
            return await func(request, *args, **kwargs)

        return wrapper

    return decorator


def get_token_info(token: dict) -> dict:
    token_info = dict()
    for key in token:
        if key != "userinfo":
            token_info[key] = token[key]
    return token_info


async def get_user_from_token(
    token: Annotated[str, Depends(oauth2_scheme)],
    request: Request,
) -> User:
    try:
        auth_provider_id = request.headers["auth_provider"]
    except KeyError:
        raise HTTPException(
            status.HTTP_401_UNAUTHORIZED,
            "Request headers must have a 'auth_provider' field",
        )
    try:
        provider = providers[auth_provider_id]
    except KeyError:
        if auth_provider_id == "":
            raise HTTPException(status.HTTP_401_UNAUTHORIZED, "No auth provider")
        else:
            raise HTTPException(
                status.HTTP_401_UNAUTHORIZED, f"Unknown auth provider '{auth_provider_id}'"
            )
    if token == "None":
        request.session.pop("auth_provider_id", None)
        request.session.pop("user_sub", None)
        raise HTTPException(status.HTTP_401_UNAUTHORIZED, "No token")
    try:
        payload = provider.decode(token)
    except ExpiredSignatureError:
        raise HTTPException(
            status.HTTP_401_UNAUTHORIZED,
            "Expired signature (token refresh not implemented yet)",
        )
    except InvalidKeyError:
        raise HTTPException(status.HTTP_401_UNAUTHORIZED, "Invalid auth provider key")
    except PyJWTError as err:
        raise HTTPException(
            status.HTTP_401_UNAUTHORIZED, f"Cannot decode token: {err.__class__.__name__}"
        )
    try:
        user_id = payload["sub"]
    except KeyError:
        return provider.unknown_auth_user
    try:
        user = await db.get_user(user_id)
        if user.access_token != token:
            user.access_token = token
    except UserNotInDB:
        logger.info(
            f"User {user_id} not found in DB, creating it (real apps can behave differently)"
        )
        user = await db.add_user(
            sub=payload["sub"],
            user_info=payload,
            auth_provider=providers[auth_provider_id],
            access_token=token,
        )
    return user


async def get_user_from_token_or_none(
    token: Annotated[str | None, Depends(oauth2_scheme_optional)],
    request: Request,
) -> User | None:
    if token is None:
        return None
    try:
        return await get_user_from_token(token, request)
    except HTTPException:
        return None


class UserWithRole:
    roles: set[str]

    def __init__(self, roles: str | list[str] | tuple[str] | set[str]):
        if isinstance(roles, str):
            self.roles = set([roles])
        elif isinstance(roles, (list, tuple, set)):
            self.roles = set(roles)

    def __call__(self, user: User = Depends(get_user_from_token)) -> User:
        if not any(self.roles.intersection(user.roles_as_set)):
            raise HTTPException(
                status.HTTP_401_UNAUTHORIZED, f"Not of any required role {', '.join(self.roles)}"
            )
        return user
