"""Fake in-memory database interface for demo purpose"""

import logging

from authlib.oauth2.rfc6749 import OAuth2Token
from jwt import PyJWTError

from oidc_test.auth.provider import Provider

from oidc_test.models import User, Role
from oidc_test.auth_providers import providers

logger = logging.getLogger("oidc-test")


class UserNotInDB(Exception):
    pass


class TokenNotInDb(Exception):
    pass


class Database:
    users: dict[str, User] = {}
    # TODO: key of the token table should be provider: sid
    tokens: dict[str, OAuth2Token] = {}

    # Last sessions for the user (key: users's subject id (sub))

    async def add_user(
        self,
        sub: str,
        user_info: dict,
        auth_provider: Provider,
        access_token: str,
        access_token_decoded: dict | None = None,
    ) -> User:
        if access_token_decoded is None:
            assert auth_provider.name is not None
            provider = providers[auth_provider.id]
            try:
                access_token_decoded = provider.decode(access_token)
            except PyJWTError:
                access_token_decoded = {}
        user_info["auth_provider_id"] = auth_provider.id
        user = User(**user_info)
        user.userinfo = user_info
        # user.access_token = access_token
        # user.access_token_decoded = access_token_decoded
        # Add roles provided in the access token
        roles = set()
        try:
            r = access_token_decoded["resource_access"][auth_provider.client_id]["roles"]
            roles.update(r)
        except KeyError:
            pass
        try:
            r = access_token_decoded["realm_access"]["roles"]
            if isinstance(r, str):
                roles.add(r)
            else:
                roles.update(r)
        except KeyError:
            pass
        user.roles = [Role(name=role_name) for role_name in roles]
        self.users[sub] = user
        return user

    async def get_user(self, sub: str) -> User:
        if sub not in self.users:
            raise UserNotInDB
        return self.users[sub]

    async def add_token(self, provider: Provider, token: OAuth2Token) -> None:
        """Store a token using as key the sid (auth provider's session id)
        in the id_token"""
        sid = provider.get_session_key(token["userinfo"])
        self.tokens[sid] = token

    async def get_token(
        self,
        provider: Provider,
        sid: str | None,
    ) -> OAuth2Token:
        # TODO: key of the token table should be provider: sid
        assert isinstance(provider, Provider)
        if sid is None:
            raise TokenNotInDb
        try:
            return self.tokens[sid]
        except KeyError:
            raise TokenNotInDb


db = Database()
