from typing import Annotated, Any
import logging
from json import JSONDecodeError

from authlib.oauth2.rfc6749 import OAuth2Token
from httpx import AsyncClient, HTTPError
from jwt.exceptions import DecodeError, ExpiredSignatureError, InvalidTokenError
from fastapi import FastAPI, HTTPException, Depends, Request, status
from fastapi.middleware.cors import CORSMiddleware

# from starlette.middleware.sessions import SessionMiddleware
# from authlib.integrations.starlette_client.apps import StarletteOAuth2App
# from authlib.oauth2.rfc6749 import OAuth2Token

from oidc_test.auth.provider import Provider
from oidc_test.auth.utils import (
    get_user_from_token_or_none,
    oauth2_scheme_optional,
)
from oidc_test.auth_providers import providers
from oidc_test.settings import ResourceProvider, settings
from oidc_test.models import User
from oidc_test.registry import ProcessError, ProcessResult, registry

logger = logging.getLogger("oidc-test")

resource_server = FastAPI()


resource_server.add_middleware(
    CORSMiddleware,
    allow_origins=settings.cors_origins,
    allow_credentials=True,
    allow_methods=["*"],
    allow_headers=["*"],
)

# SessionMiddleware is required by authlib
# resource_server.add_middleware(
#    SessionMiddleware,
#    secret_key=settings.secret_key,
# )

# Route for OAuth resource server


# Routes for RBAC based tests


@resource_server.get("/")
async def resources() -> dict[str, dict[str, Any]]:
    return {"internal": {}, "plugins": registry.resources}


@resource_server.get("/{resource_name}")
@resource_server.get("/{resource_name}/{resource_id}")
async def get_resource(
    resource_name: str,
    user: Annotated[User | None, Depends(get_user_from_token_or_none)],
    token: Annotated[OAuth2Token | None, Depends(oauth2_scheme_optional)],
    resource_id: str | None = None,
):
    """Generic path for testing a resource provided by a provider.
    There's no field validation (response type of ProcessResult) on purpose,
    leaving the responsibility of the response validation to resource providers"""
    # Get the resource if it's defined in user auth provider's resources (external)
    if user is not None:
        provider = providers[user.auth_provider_id]
        if ":" in resource_name:
            # Third-party resource provider: send the request with the request token
            resource_provider_id, resource_name = resource_name.split(":", 1)
            provider = providers[user.auth_provider_id]
            resource_provider: ResourceProvider = provider.get_resource_provider(
                resource_provider_id
            )
            resource_url = resource_provider.get_resource_url(resource_name)
            async with AsyncClient(verify=resource_provider.verify_ssl) as client:
                try:
                    logger.debug(f"GET request to {resource_url}")
                    resp = await client.get(
                        resource_url,
                        headers={
                            "Content-type": "application/json",
                            "Authorization": f"Bearer {token}",
                            "auth_provider": user.auth_provider_id,
                        },
                    )
                except HTTPError as err:
                    raise HTTPException(
                        status.HTTP_503_SERVICE_UNAVAILABLE, err.__class__.__name__
                    )
                except Exception as err:
                    raise HTTPException(
                        status.HTTP_500_INTERNAL_SERVER_ERROR, err.__class__.__name__
                    )
                else:
                    if resp.is_success:
                        return resp.json()
                    else:
                        reason_str: str
                        try:
                            reason_str = resp.json().get("detail", str(resp))
                        except Exception:
                            reason_str = str(resp.text)
                        raise HTTPException(resp.status_code, reason_str)
        # Third party resource (provided through the auth provider)
        # The token is just passed on
        # XXX: is this branch valid anymore?
        if resource_name in [r.resource_name for r in provider.resources]:
            return await get_auth_provider_resource(
                provider=provider,
                resource_name=resource_name,
                token=token,
                user=user,
            )
    # Internal resource (provided here)
    if resource_name in registry.resources:
        resource = registry.resources[resource_name]
        reason: dict[str, str] = {}
        if not resource.is_public:
            if user is None:
                raise HTTPException(status.HTTP_401_UNAUTHORIZED, "Resource is not public")
            else:
                if resource.scope_required is not None and not user.has_scope(
                    resource.scope_required
                ):
                    reason["scope"] = (
                        f"No scope {resource.scope_required} in the access token "
                        "but it is required for accessing this resource"
                    )
                if (
                    resource.role_required is not None
                    and resource.role_required not in user.roles_as_set
                ):
                    reason["role"] = (
                        f"You don't have the role {resource.role_required} "
                        "but it is required for accessing this resource"
                    )
        if len(reason) == 0:
            try:
                resp = await resource.process(user=user, resource_id=resource_id)
                return resp
            except ProcessError as err:
                raise HTTPException(
                    status.HTTP_401_UNAUTHORIZED, f"Cannot process resource: {err}"
                )
        else:
            raise HTTPException(status.HTTP_401_UNAUTHORIZED, ", ".join(reason.values()))
    else:
        raise HTTPException(status.HTTP_401_UNAUTHORIZED, "Unknown resource")
    # return await get_resource_(resource_name, user, **request.query_params)


async def get_auth_provider_resource(
    provider: Provider, resource_name: str, token: OAuth2Token | None, user: User
) -> ProcessResult:
    if token is None:
        raise HTTPException(status.HTTP_401_UNAUTHORIZED, "No auth token")
    access_token = token
    async with AsyncClient() as client:
        resp = await client.get(
            url=provider.get_resource_url(resource_name),
            headers={
                "Content-type": "application/json",
                "Authorization": f"Bearer {access_token}",
            },
        )
    if resp.is_error:
        raise HTTPException(resp.status_code, f"Cannot fetch resource: {resp.reason_phrase}")
    # Only a demo, real application would really process the response
    resp_length = len(resp.text)
    if resp_length > 1024:
        return ProcessResult(
            msg=f"The resource is too long ({resp_length} bytes) to show in this demo, here is just the begining in raw format",
            start=resp.text[:100] + "...",
        )
    else:
        try:
            resp_json = resp.json()
        except JSONDecodeError:
            return ProcessResult(msg="The resource is not formatted in JSON", text=resp.text)
        if isinstance(resp_json, dict):
            return ProcessResult(**resp.json())
        elif isinstance(resp_json, list):
            return ProcessResult(**{str(i): line for i, line in enumerate(resp_json)})


# @resource_server.get("/public")
# async def public() -> dict:
#    return {"msg": "Not protected"}
#
#
# @resource_server.get("/protected")
# async def get_protected(user: Annotated[User, Depends(get_user_from_token)]):
#    assert user is not None  # Just to keep QA checks happy
#    return {"msg": "Only authenticated users can see this"}
#
#
# @resource_server.get("/protected-by-foorole")
# async def get_protected_by_foorole(
#    user: Annotated[User, Depends(UserWithRole("foorole"))],
# ):
#    assert user is not None
#    return {"msg": "Only users with foorole can see this"}
#
#
# @resource_server.get("/protected-by-barrole")
# async def get_protected_by_barrole(
#    user: Annotated[User, Depends(UserWithRole("barrole"))],
# ):
#    assert user is not None
#    return {"msg": "Protected by barrole"}
#
#
# @resource_server.get("/protected-by-foorole-and-barrole")
# async def get_protected_by_foorole_and_barrole(
#    user: Annotated[User, Depends(UserWithRole("foorole")), Depends(UserWithRole("barrole"))],
# ):
#    assert user is not None  # Just to keep QA checks happy
#    return {"msg": "Only users with foorole and barrole can see this"}
#
#
# @resource_server.get("/protected-by-foorole-or-barrole")
# async def get_protected_by_foorole_or_barrole(
#    user: Annotated[User, Depends(UserWithRole(["foorole", "barrole"]))],
# ):
#    assert user is not None  # Just to keep QA checks happy
#    return {"msg": "Only users with foorole or barrole can see this"}

# async def get_resource_(resource_id: str, user: User, **kwargs) -> dict:
#    """
#    Resource processing: build an informative rely as a simple showcase
#    """
#    if resource_id == "petition":
#        return await sign(user, kwargs["petition_id"])
#    provider = providers[user.auth_provider_id]
#    try:
#        pname = provider.name
#    except KeyError:
#        pname = "?"
#    resp = {
#        "hello": f"Hi {user.name} from an OAuth resource provider",
#        "comment": f"I received a request for '{resource_id}' "
#        + f"with an access token signed by {pname}",
#    }
#    # For the demo, resource resource_id matches a scope get:resource_id,
#    # but this has to be refined for production
#    required_scope = f"get:{resource_id}"
#    # Check if the required scope is in the scopes allowed in userinfo
#    try:
#        if user.has_scope(required_scope):
#            await process(user, resource_id, resp)
#        else:
#            ## For the showcase, giving a explanation.
#            ## Alternatively, raise HTTP_401_UNAUTHORIZED
#            raise HTTPException(
#                status.HTTP_401_UNAUTHORIZED,
#                f"No scope {required_scope} in the access token "
#                + "but it is required for accessing this resource",
#            )
#    except ExpiredSignatureError:
#        raise HTTPException(status.HTTP_401_UNAUTHORIZED, "The token's signature has expired")
#    except InvalidTokenError:
#        raise HTTPException(status.HTTP_401_UNAUTHORIZED, "The token is invalid")
#    return resp


# async def process(user, resource_id, resp):
#    """
#    Too simple to be serious.
#    It's a good fit for a plugin architecture for production
#    """
#    if resource_id == "time":
#        resp["time"] = datetime.now().strftime("%c")
#    elif resource_id == "bs":
#        async with AsyncClient() as client:
#            bs = await client.get("https://corporatebs-generator.sameerkumar.website/")
#            resp["bs"] = bs.json().get("phrase", "Sorry, i am out of BS today.")
#    else:
#        raise HTTPException(
#            status.HTTP_401_UNAUTHORIZED, f"I don't known how to give '{resource_id}'."
#        )


# @resource_server.get("/introspect")
# async def get_introspect(
#    request: Request,
#    oidc_provider: Annotated[StarletteOAuth2App, Depends(get_oidc_provider)],
#    token: Annotated[OAuth2Token, Depends(get_token)],
# ) -> JSONResponse:
#    assert request is not None  # Just to keep QA checks happy
#    if (url := oidc_provider.server_metadata.get("introspection_endpoint")) is None:
#        raise HTTPException(
#            status_code=status.HTTP_401_UNAUTHORIZED,
#            detail="No introspection endpoint found for the OIDC provider",
#        )
#    if (
#        response := await oidc_provider.post(
#            url,
#            token=token,
#            data={"token": token["access_token"]},
#        )
#    ).is_success:
#        return response.json()
#    else:
#        raise HTTPException(status_code=response.status_code, detail=response.text)

# assert user.oidc_provider is not None
### Get some info (TODO: refactor)
# if (auth_provider_id := user.oidc_provider.name) is None:
#    raise HTTPException(
#        status.HTTP_401_UNAUTHORIZED,
#        "Request headers must have a 'auth_provider' field",
#    )
# if (
#    auth_provider_settings := oidc_providers_settings.get(auth_provider_id)
# ) is None:
#    raise HTTPException(
#        status.HTTP_401_UNAUTHORIZED, f"Unknown auth provider '{auth_provider_id}'"
#    )
# if (key := auth_provider_settings.get_public_key()) is None:
#    raise HTTPException(
#        status.HTTP_401_UNAUTHORIZED,
#        f"Key for provider '{auth_provider_id}' unknown",
#    )
# logger.warn(f"refresh with scope {scope}")
# breakpoint()
# refreshed_auth_info = await user.oidc_provider.fetch_access_token(scope=scope)
### Decode the new token
# try:
#    payload = decode(
#        refreshed_auth_info["access_token"],
#        key=key,
#        algorithms=["RS256"],
#        audience="account",
#        options={"verify_signature": not settings.insecure.skip_verify_signature},
#    )
# except ExpiredSignatureError as err:
#    logger.info(f"Expired signature: {err}")
#    raise HTTPException(
#        status.HTTP_401_UNAUTHORIZED,
#        "Expired signature (refresh not implemented yet)",
#    )
